#####################################################################
# 
# ZTxnQueue.py - Zope Transactional Queue 
#
# $Id: ZTxnQueue.py,v 1.1 2003/11/19 18:38:43 kevin Exp $
#
# Copyright (C) 2003 Kevin Gill
#
# Derived from the Xron version 0.0.10
# Copyright (C) 2000 Loren Stafford
#  
# Xron is Derived from the ZScheduler product version 0.0.7 in
# accord with the terms of the ZScheduler's license.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License  
# along with this program; if not, write to the 
#     Free Software Foundation, Inc., 
#     59 Temple Place - Suite 330, 
#     Boston, MA  02111-1307, USA.
# 
#####################################################################

""" Zope Transactional Queue """

import string
import TQInterface, TQPollingThread
from TQUtility import UniqueId

# I use the thread lock to ensure that only one thread
# is dequeuing at a time. Otherwise the same values could
# be passed to different transactions
import ThreadLock
_mutex = ThreadLock.allocate_lock()

from Interface import Base
from ZODB.PersistentList import PersistentList
from Persistence import Persistent
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from AccessControl import ClassSecurityInfo, getSecurityManager, Role
from OFS.SimpleItem import Item
from OFS.PropertyManager import PropertyManager
from Globals import InitializeClass
import Acquisition
from Shared.DC.ZRDB.TM import TM

import zLOG

# I want to track dequeues - Multiple threads access the queue
# they will be blocked, being forced to access it serially. However,
# each instance of a blocked class will have stale data.
# _TrackedDequeues is a centralised store, where a stale
# object can identify the point up to which items have been dequeued.
# 
# The Last dequeued object id is here.
# If that is in your version of the data, simply wind forward
# to it. 
#
# I use the physical path to the zope object as an index to this
# file. This, I believe is inefficient and I also believe that
# there must be some better id around for this. I cannot use
# 'id' as it may be duplicated in a different folder. I cannot
# generate a key, because the object could be copied through ZMI.
_TrackedDequeues={}


class ZTxnQueue(TM, PersistentList, Item, PropertyManager, Acquisition.Implicit, Role.RoleManager):
    """Cheap and dirty implementation. I use a persistent mapping to store
    the queue, and the keys are sequential to maintain the queue ordering."""

    def enqueue(self, Data):
        """Enqueues a piece of data. """
        return self._enqueue(Data)
    def dequeue(self):
        """Pops an item off the queue. """
        return self._dequeue()
    def items(self):
        """View the Contents of the Queue - For the Management Screens"""
        self._AmIManaged()
        rv = []
        for i in self:
            rv.append(i)
        return rv
    def find(self, OID):
        """Returns the data for OID or None if not found"""
        for (storedOID, storedData) in self:
            if storedOID == OID:
                return storedData
        return None
    def __init__(self, id, title):
        """Initialise """
        self.id = id
        self.path = None
        self.title = title
        PersistentList.__init__(self)

        # These are the things to enable a queue proccessing function
        self.ProcessingState='A' # A=Active, B=Blocked
        self.ProcessingObject=None
        self.QueueProcessSize=10

    def sleep(self, seconds):
        """I just put this in so I could test concurrency easily"""
        import time
        time.sleep(seconds)

    ### Consider everything beyond this private

    __ac_roles__ = ()
    __implements__ = (TQInterface.IZTxnQueue, TQInterface.IPolledObject)
    meta_type = 'ZTxnQueue'
    security = ClassSecurityInfo()

    security.declareProtected('ZTxnQueue Modify', 'enqueue')
    security.declareProtected('ZTxnQueue Modify', 'dequeue')
    security.declarePublic('items')
    security.declarePublic('find')
    security.setPermissionDefault('ZTxnQueue Modify', ('Manager', 'Owner'))

    manage_options = ((
       {'label':'Contents', 'action':'manage_main', 'help': ('ZTxnQueue', 'AZTxnQueue.stx')},)+ 
       PropertyManager.manage_options+Role.RoleManager.manage_options+ Item.manage_options)
        
    manage_main = PageTemplateFile('www/manage_View.zpt', globals())
    _properties = ({'id': 'title', 'type': 'string', 'mode': 'w'},
                   {'id': 'path', 'type': 'string', 'mode': 'r'},
                   {'id': 'ProcessingState', 'type': 'string', 'mode': 'w'},
                   {'id': 'ProcessingObject', 'type': 'string', 'mode': 'w'},
                   {'id': 'QueueProcessSize', 'type': 'string', 'mode': 'w'},
    )

    security.declarePrivate('_enqueue')
    def _enqueue(self, Data):
        """Place an item into the queue"""
        if not hasattr(self, '_v_registered') or get_transaction()._id not in self._v_registered:
            self._registered = 0  # There seems to be two registrations going required.
            self._register()
            get_transaction().register(self)
        self._p_changed = 1
        OID = UniqueId()
        self.append((OID, Data))
        return OID

    security.declarePrivate('_dequeue')
    def _dequeue(self):
        """Dequeue an item"""
        if not hasattr(self, '_v_gotlock') or self._v_gotlock == 'N':
            _mutex.acquire()
            self._v_gotlock = 'Y'
             # Where this object is older than the database version,
            # it will have items at the start that are already
            # processed - wind past these.
            try:
                lastDequeued = _TrackedDequeues[self.path]
            except:
                lastDequeued = None
            self._v_initialDequeued = lastDequeued
            if lastDequeued:
                index = 0
                for (oid, data) in self:
                    index+=1
                    if lastDequeued == oid:
                       del self[:index] # Remove the items up to / including oid
                       self._p_changed = 1
                       break
            if not hasattr(self, '_v_registered') or get_transaction()._id not in self._v_registered:
                # Need to register both for TM and Persistence
                self._registered = 0  # since not _v_ it get stored!!!
                self._register()                   
                get_transaction().register(self)
        # Finally, down to business
        try:
            rv = self.pop(0)
        except:
            rv = None
        if rv:
            _TrackedDequeues[self.path] = rv[0]
            self._p_changed = 1
        return rv

    def _p_resolveConflict(self, oldState, savedState, newState):
        """This module contains a variable which tracks the last
        dequeued item. If we assume that OIDs are sequential, we
        can say that the resulting queue from merging the newState
        and the savedState must be everything with an OID
        greater than the lastDequeued item. - Of course this is
        except in the case of an abort during dequeuing."""
        if len(oldState['data']) > 0:
            zLOG.LOG('ZTxnQueue', zLOG.BLATHER, "oldState %d [%s ... %s]" %
                (len(oldState['data']), oldState['data'][0][0], oldState['data'][-1][0]))
        else:
            zLOG.LOG('ZTxnQueue', zLOG.BLATHER, "oldState is empty") 
        if len(newState['data']) > 0:
            zLOG.LOG('ZTxnQueue', zLOG.BLATHER, "newState %d [%s ... %s]" % 
                (len(newState['data']), newState['data'][0][0], newState['data'][-1][0]))
        else:
            zLOG.LOG('ZTxnQueue', zLOG.BLATHER, "newState is empty") 
        if len(savedState['data']) > 0:
            zLOG.LOG('ZTxnQueue', zLOG.BLATHER, "savedState %d [%s ... %s]" % 
                (len(savedState['data']), savedState['data'][0][0], savedState['data'][-1][0]))
        else:
            zLOG.LOG('ZTxnQueue', zLOG.BLATHER, "savedState is empty") 
        zLOG.LOG('ZTxnQueue', zLOG.BLATHER, "_TrackedDequeues [%s]" % (_TrackedDequeues))
        zLOG.LOG('ZTxnQueue', zLOG.BLATHER, "PhysicalPath [%s]" % (newState['path']))

        try:
            lastOID = _TrackedDequeues[newState['path']]
        except:
            lastOID = 0L

        mergedData = []
        (savedIdx, newIdx) = (0, 0)
        (savedData, newData) = (savedState['data'], newState['data'])
        while 1:
            if savedIdx >= len(savedData):
                ### We have reached the end of savedData, Dump
                ### the rest of newData onto the result and return
                mergedData = mergedData + newData[newIdx:]
                break
            if newIdx >= len(newData):
                ### We have reached the end of newData. Dump
                ### the rest of savedData to the result and return
                mergedData = mergedData + savedData[newIdx:]
                break

            if savedData[savedIdx][0] <= lastOID:
                savedIdx += 1
                continue
            if newData[newIdx][0] <= lastOID:
                newIdx += 1
                continue
            if newData[newIdx][0] <= savedData[savedIdx][0]:
                mergedData.append(newData[newIdx])
                lastOID = newData[newIdx][0]
                newIdx += 1
            else:
                mergedData.append(savedData[savedIdx])
                lastOID = savedData[savedIdx][0]
                savedIdx += 1

        if len(mergedData) > 0:
            zLOG.LOG('ZTxnQueue', zLOG.BLATHER, "Result %d [%s ... %s]" %  (len(mergedData), mergedData[0][0], mergedData[-1][0]))
        else:
            zLOG.LOG('ZTxnQueue', zLOG.BLATHER, "Result is empty")
        ## Must use the savedState so that new versions of properties are
        ## included
        savedState['data'] = mergedData
        return oldState


    # I am using a MUTEX to ensure that only one
    # threads can get dequeue access to the object at a time.
    # If two threads started dequeuing, they would only 
    # find out after it was too late - i.e. on a commit() !! 
    # The _registered flag seems to get pickled under certain
    # situations, leading to transactions being disabled. I 
    # have a shadow variable _v_registered.
    def _begin(self):
        """Begin"""
        _id =  get_transaction()._id
        if not hasattr(self, '_v_registered'): self._v_registered = []
        if not hasattr(self, '_v_gotlock'): self._v_gotlock = 'N'
        #zLOG.LOG('ZTxnQueue', zLOG.WARNING, "begin(), [%s], %s, txn %s, _v_registered = %s, _v_gotlock = %s" %
        #    (self.path, id(self), _id, self._v_registered, self._v_gotlock))
        if _id not in self._v_registered: self._v_registered.append(_id)
    def _finish(self):
        """Commit"""
        _id =  get_transaction()._id
        if not hasattr(self, '_v_registered'): self._v_registered = []
        if not hasattr(self, '_v_gotlock'): self._v_gotlock = 'N'
        #zLOG.LOG('ZTxnQueue', zLOG.WARNING, "finish(), [%s], %s, txn %s, _v_registered = %s, _v_gotlock = %s" %
        #    (self.path, id(self), _id, self._v_registered, self._v_gotlock))
        try:
            # There is some bizare behaviour in the case of an abort(). It seems to 
            # be called using an object which has lost _v_ variables
            _mutex.release()
        except:
            pass
        self._v_gotlock = 'N'
        self._v_initialDequeued = None
        try:
            self._v_registered.remove(_id)
        except:
            pass
    def _abort(self):
        """Rollback"""
        _id =  get_transaction()._id
        if not hasattr(self, '_v_registered'): self._v_registered = []
        if not hasattr(self, '_v_gotlock'): self._v_gotlock = 'N'
        #zLOG.LOG('ZTxnQueue', zLOG.WARNING, "abort(), [%s], %s, txn %s, _v_registered = %s, _v_gotlock = %s" %
        #    (self.path, id(self), _id, self._v_registered, self._v_gotlock))
        if hasattr(self, '_v_initialDequeud') and self._v_initialDequeued != None:
            _TrackedDequeues[self.path] = self._v_initialDequeued
        try:
            # There is some bizare behaviour in the case of an abort(). It seems to 
            # be called using an object which has lost _v_ variables
            _mutex.release()
        except:
            pass
        self._v_gotlock = 'N'
        self._v_initialDequeued = None
        try:
            self._v_registered.remove(_id)
        except:
            pass

    def _AmIManaged(self):
        """Verify that the current instance is managed. Also handle renames"""
        if hasattr(self, '_v_managed'): return
        ManagerObj = getattr(self, TQPollingThread.GlobalID, None)
        if not ManagerObj: return
        self._v_managed = 'Y'
        current_path = string.join(self.getPhysicalPath(), "/")
        if not self.path:
            self.path = current_path
            self._p_changed = 1
            ManagerObj.registerObject(self)
        elif self.path != current_path:
            old_path = self.path
            self.path = current_path
            self._p_changed = 1
            ManagerObj.reregisterObject(old_path, self)
        else:
            ManagerObj.registerObject(self)

    def manage_beforeDelete(self, item, container):
        """Called when the queue is deleted - remove Polling Thread entry"""
        if not self.path:
            return
        ManagerObj = getattr(self, TQPollingThread.GlobalID, None)
        if not ManagerObj: return
        ManagerObj.unregisterObject(self)

    def Process(self, App, private_data):
        """This method is called every 10 seconds (configurable) to
        process the queue stored in this object."""

        if self.ProcessingState != 'A': return 0   # Inactive
        if len(self.data) == 0: return 0           # Nothing to Do

        # Processing Queue Commences Here
        ProcessingObjectPath = self.ProcessingObject
        if ProcessingObjectPath == None or ProcessingObjectPath == '': return 0  # No processing object
        try:
            if ProcessingObjectPath[0] == '/':
                ProcessingObject = App.unrestrictedTraverse(ProcessingObjectPath)
            else:
                ProcessingObject = self.unrestrictedTraverse(ProcessingObjectPath)
        except:
            import sys, traceback
            (type, val, tb) = sys.exc_info()
            zLOG.LOG('ZTxnQueue', zLOG.PANIC,
                "ZTxnQueue %s - Processing Object %s Exception\n%s %s\n%s\n" %
                (self.path, ProcessingObjectPath, type, val, 
                string.join(traceback.format_exception(type, val, tb), '\n')))
            get_transaction().abort()
            # Cannot instanciate the Processing Object - Block the queue
            get_transaction().begin()
            self.ProcessingState = 'B'
            self._p_changed = 1
            get_transaction().commit()
            get_transaction().begin()
            return 0 

        try:
            QueueProcessSize = int(self.QueueProcessSize)
        except:
            zLOG.LOG('ZTxnQueue', zLOG.WARNING,
                "ZTxnQueue %s - Has invalid QueueProcessSize %s - Setting to 10" %
                (self.path, self.QueueProcessSize))
            QueueProcessSize = 10
            self.QueueProcessSize = 10
            self._p_changed = 1
            get_transaction().commit()
            get_transaction().begin()

        (oid, data) = (None, None)
        try:
            for idx in range(QueueProcessSize):
                rv = self.dequeue()
                if not rv: break
                (oid, data) = rv
                rvMessage = ProcessingObject(Data=data, OID=oid)
                if rvMessage != None:
                    # If anything is returned, just log it
                    zLOG.LOG('ZTxnQueue', zLOG.BLATHER, str(rvMessage))
            if oid != None:
                get_transaction().commit()
                get_transaction().begin()
            if rv:
                return 1
        except:
            import sys, traceback
            (type, val, tb) = sys.exc_info()
            zLOG.LOG('PollingThread', zLOG.ERROR, "ZTxnQueue %s OID %s - Exception Occurred\n%s %s\n%s\n" %
                (self.path, oid, type, val,
                string.join(traceback.format_exception(type, val, tb), '\n')))

            # Wind the Tracked Dequeued back to the start in case of a conflict resolution
            _TrackedDequeues[self.path] = self._v_initialDequeued
            self._v_initialDequeued = None
            get_transaction().abort()
            get_transaction().begin()
            self.ProcessingState = 'B'
            self._p_changed = 1
            get_transaction().commit()
            get_transaction().begin()
        return 0

    def getStatusMessage(self):
        if self.ProcessingState == 'B':
            return "<b><font color=red>Blocked</font></b>&nbsp;&nbsp;Queue Size = %d" % len(self.data)
        else:
            return "Active&nbsp;&nbsp;Queue Size = %d" % len(self.data)

InitializeClass(ZTxnQueue)

manage_addZTxnQueueForm = PageTemplateFile('www/manage_addZTxnQueueForm.zpt', globals())

def manage_addZTxnQueue(dispatcher, id, title='', REQUEST=None):
    """Adds a new ZTxnQeueue object with id *id*.
    """
    id = str(id)
    ob = ZTxnQueue(id, title)
    dispatcher._setObject(id, ob)
    ob = dispatcher._getOb(id)
    # Register with the TQPollingThread
    ob._AmIManaged()
    if REQUEST is not None:
        return dispatcher.manage_main(dispatcher, REQUEST)

